
//		Ancestry.cpp


#ifndef ANCESTRY_CPP_VER
#define ANCESTRY_CPP_VER				0.99.0.0

#include <tlhelp32.h>



unsigned int GetAnc_CountProcesses( HANDLE Snapshot )
{
	if ( !Snapshot )
		return 0;


	PROCESSENTRY32W ProcessInfo;

	unsigned int n				= 0;


	memset( &ProcessInfo, 0x00, sizeof( PROCESSENTRY32W ) );
	ProcessInfo.dwSize = sizeof( PROCESSENTRY32W );

	if ( Process32FirstW( Snapshot, &ProcessInfo ) ) {
		n++;
		memset( &ProcessInfo, 0x00, sizeof( PROCESSENTRY32W ) );
		ProcessInfo.dwSize = sizeof( PROCESSENTRY32W );

		while ( Process32NextW( Snapshot, &ProcessInfo ) ) {
			n++;
			memset( &ProcessInfo, 0x00, sizeof( PROCESSENTRY32W ) );
			ProcessInfo.dwSize = sizeof( PROCESSENTRY32W );
		}
	}

	return n;
}



PROCESSENTRY32W *GetAnc_FindProcess( PROCESSENTRY32W *ProcList, unsigned int NumProcesses, DWORD Pid )
{
	if ( !ProcList || !NumProcesses || !Pid )
		return NULL;


	for ( unsigned int i = 0; i < NumProcesses; i++ )
		if ( ProcList[i].th32ProcessID == Pid )
			return ProcList + i;

	return NULL;
}



int GetAnc_GetStartTime( DWORD Pid, FILETIME *StartTime )
{
	if ( !Pid || !StartTime )
		return -1;


	DWORD Access			= ( OSVerInfo.dwMajorVersion > 5 ) ? PROCESS_QUERY_LIMITED_INFORMATION : PROCESS_QUERY_INFORMATION;
	HANDLE ProcessHandle	= NULL;
	FILETIME ExitTime		= { 0, 0 };
	FILETIME KernelTime		= { 0, 0 };
	FILETIME UserTime		= { 0, 0 };
	int rv					= 0;


	StartTime->dwHighDateTime = StartTime->dwLowDateTime = 0;

	ProcessHandle = OpenProcess( Access, FALSE, Pid );
	if ( ProcessHandle ) {
		if ( GetProcessTimes( ProcessHandle, StartTime, &ExitTime, &KernelTime, &UserTime ) )
			rv = 1;

		CloseHandle( ProcessHandle );
	}

	return rv;
}



int GetAnc_CmpTimes( FILETIME Time1, FILETIME Time2 )
{
	if ( Time1.dwHighDateTime > Time2.dwHighDateTime )
		return 1;
	else if ( Time1.dwHighDateTime < Time2.dwHighDateTime )
		return -1;
	else if ( Time1.dwLowDateTime > Time2.dwLowDateTime )
		return 1;
	else if ( Time1.dwLowDateTime < Time2.dwLowDateTime )
		return -1;
	else
		return 0;
}



int GetAncestors( DWORD Pid, DWORD **Ancestors, unsigned int *NumAncestors )
{
	if ( !Ancestors || !NumAncestors )
		return -1;

	if ( !Pid )
		Pid = GetCurrentProcessId();

	if ( !Pid )
		return -1;

	if ( *Ancestors ) {
		free( *Ancestors );
		*Ancestors = NULL;
	}

	*NumAncestors = 0;



	HANDLE Snapshot					= NULL;
	unsigned int NumProcesses		= 0;
	PROCESSENTRY32W *ProcList		= NULL;
	PROCESSENTRY32W ProcessInfo;
	int i							= 0;
	unsigned int na					= 0;
	DWORD ThisPid					= Pid;
	PROCESSENTRY32W *ThisProcess	= NULL;
	DWORD *AncList					= NULL;
	FILETIME StartTime				= { 0, 0 };
	FILETIME ParentStartTime		= { 0, 0 };
	DWORD *outList					= NULL;

	int rv							= 0;


	//	wprintf( L"   *  Get ancestry for PID 0x%08X:\n", Pid );

	Snapshot = CreateToolhelp32Snapshot( TH32CS_SNAPPROCESS, NULL );
	if ( Snapshot ) {
		//	wprintf( L"   *  Got snapshot.\n" );

		NumProcesses = GetAnc_CountProcesses( Snapshot );
		//	wprintf( L"   *  %u processes.\n", NumProcesses );

		ProcList = (PROCESSENTRY32W*) malloc( NumProcesses * sizeof( PROCESSENTRY32W ) );
		if ( ProcList ) {
			//	wprintf( L"   *  Allocated buffer.\n" );

			memset( ProcList, 0x00, NumProcesses * sizeof( PROCESSENTRY32W ) );

			i = 0;
			memset( &ProcessInfo, 0x00, sizeof( PROCESSENTRY32W ) );
			ProcessInfo.dwSize = sizeof( PROCESSENTRY32W );

			if ( Process32FirstW( Snapshot, &ProcessInfo ) ) {
				ProcList[i] = ProcessInfo;
				i++;

				memset( &ProcessInfo, 0x00, sizeof( PROCESSENTRY32W ) );
				ProcessInfo.dwSize = sizeof( PROCESSENTRY32W );

				while ( Process32NextW( Snapshot, &ProcessInfo ) ) {
					ProcList[i] = ProcessInfo;
					i++;

					memset( &ProcessInfo, 0x00, sizeof( PROCESSENTRY32W ) );
					ProcessInfo.dwSize = sizeof( PROCESSENTRY32W );
				}
			}

			//	wprintf( L"   *  Slurped info for %u processes.\n", i );

			AncList = (DWORD*) malloc( NumProcesses * sizeof( DWORD ) );
			if ( AncList ) {
				//	wprintf( L"   *  Allocated temp ancestors buffer.\n" );

				na = 0;
				ThisPid = Pid;
				ThisProcess = GetAnc_FindProcess( ProcList, NumProcesses, ThisPid );

				if ( ThisProcess ) {
					GetAnc_GetStartTime( ThisPid, &StartTime );
					//	wprintf( L"   *  AncList[%u] = 0x%08X    '%s'\n", na, ThisProcess->th32ProcessID, ThisProcess->szExeFile );
					AncList[na]	= ThisPid;
					na++;

					while ( ThisProcess->th32ParentProcessID ) {
						ThisPid = ThisProcess->th32ParentProcessID;

						ThisProcess = GetAnc_FindProcess( ProcList, NumProcesses, ThisPid );
						if ( !ThisProcess )
							break;

						if ( StartTime.dwHighDateTime || StartTime.dwLowDateTime ) {
							GetAnc_GetStartTime( ThisPid, &ParentStartTime );
							if ( ParentStartTime.dwHighDateTime	|| ParentStartTime.dwLowDateTime )
								if ( GetAnc_CmpTimes( StartTime, ParentStartTime ) < 0 )
									break;
						}

						//	wprintf( L"   *  AncList[%u] = 0x%08X    '%s'\n", na, ThisProcess->th32ProcessID, ThisProcess->szExeFile );
						AncList[na] = ThisPid;
						na++;

						StartTime = ParentStartTime;
					}

					//	wprintf( L"   *  Ancestors found:  %u\n", na );
				}

				if ( na ) {		//	succeeded
					outList			= (DWORD*) malloc( na * sizeof( DWORD ) );

					if ( outList ) {
						for ( unsigned int i = 0; i < na; i++ )
							outList[i] = AncList[i];

						*Ancestors		= outList;
						*NumAncestors	= na;
						rv = na;
					}
				}

				free( AncList );
			}

			free( ProcList );
		}

		CloseHandle( Snapshot );
	}

	return rv;
}



int GetMyAncestors( DWORD **Ancestors, unsigned int *NumAncestors )
{
	return GetAncestors( GetCurrentProcessId(), Ancestors, NumAncestors );
}


#endif	//	ANCESTRY_CPP_VER


/*

	0.99.0.0		2024-04-29

*/
