
//			PopupFix.cpp  --  plugin to ensure that TCC's popup dialog boxes are visible


#define UNICODE						1
#define _UNICODE					1

#include <stdio.h>
#include <tchar.h>

#include <windows.h>
#include <Psapi.h>


#include "PlugIn.h"
#include "TakeCmd.h"


#define VER_MAJOR					1
#define VER_MINOR					4
#define VER_BUILD					2
#define VER_PATCH					0

#define PLUGIN_NAME					L"PopupFix"

#ifdef _WIN64
#define PLATFORM					L"x64"
#else
#define PLATFORM					L"x86"
#endif


#define MAX_ARG_LEN					8192
#define MAX_ARGS					1024
#define SHORT_BUF_LEN				1024
#define MAX_FILENAME_LEN			16384

#define MAX_POPUP_WINDOWS			32
#define MAX_CONSOLE2_WINDOWS		32
#define MAX_WINTERM_WINDOWS			32

#define TIMER_MS_MIN				10
#define TIMER_MS_MAX				1000
#define TIMER_MS_DEFAULT			125
#define RES_CHECK_SECONDS			3

#define RESTORE_FOCUS_NEVER			0
#define RESTORE_FOCUS_ALWAYS		1
#define RESTORE_FOCUS_CONSOLE2		2
#define RESTORE_FOCUS_MAX_VALUE		2

#define CENTER_AS_NEEDED			0
#define CENTER_ALWAYS				1
#define CENTER_REPOSITION			2

#define ERRHANDLE					GetStdHandle( STD_ERROR_HANDLE )

#define MAGIC_BYPASS_VALUE			0xfedcba98



wchar_t Features[] = L"popupfix,_conemu,_console2,_tcmdtab,_winterm";




const wchar_t PopupClass[]					= L"TakeCommandListBox";
const wchar_t DialogClass[]					= L"#32770";
const wchar_t TCmdClass[]					= L"Afx:00400000:8:";
const wchar_t TCmdClass2[]					= L"XTPMainFrame";
const int TCmdClassLen						= 15;
const wchar_t Console2Class[]				= L"Console_2_Main";
const wchar_t ConEmuClass[]					= L"VirtualConsoleClass";


const wchar_t PluginKeyName[]				= L"Software\\JPPlugins\\PopupFix";

const wchar_t AlwaysCenterRegName[]			= L"AlwaysCenter";
const wchar_t RestoreFocusRegName[]			= L"RestoreFocus";
const wchar_t Console2OnlyRegName[]			= L"Console2Only";
const wchar_t TimerMSRegName[]				= L"TimerMS";

const wchar_t AlwaysCenterEnvName[]			= L"PopupFix_Always_Center";
const wchar_t RestoreFocusEnvName[]			= L"PopupFix_Restore_Focus";
const wchar_t Console2OnlyEnvName[]			= L"PopupFix_Console2_Only";
const wchar_t TimerMSEnvName[]				= L"PopupFix_Timer";



const wchar_t HelpWindowTitle[]				= L"PopupFix Plugin Help";



//		--------------------  Global Variables  --------------------


DWORD MyThread						= NULL;

OSVERSIONINFOEXW OSVerInfo;

DWORD *MyAncestors					= NULL;
unsigned int NumAncestors			= 0;

HWND PopupWindowHandle[MAX_POPUP_WINDOWS];

UINT_PTR TimerID					= 0;
int TimerMS							= 0;
int ResCheckCycles					= 0;
int ResCheckCount					= 0;

int ScreenLeft = 0, ScreenTop = 0, ScreenWidth = 0, ScreenHeight = 0;
int ScreenRight = 0, ScreenBottom = 0;


HWND Console2WindowHandle			= NULL;
HWND TCmdWindowHandle				= NULL;
HWND ConEmuWindowHandle				= NULL;
HWND WinTermWindowHandle			= NULL;
HWND ConsoleWindowHandle			= NULL;
int ConsoleWindowVisible			= -1;


int CenterPopups					= CENTER_ALWAYS;
int RestoreFocus					= RESTORE_FOCUS_CONSOLE2;
int Console2Only					= FALSE;

BOOL RefocusFlag					= FALSE;
BOOL NewPopupsFound					= FALSE;


//		------------------------------------------------------------



const wchar_t PopupFixHelpText[] = L"Display or change PopupFix options.\n\
\n\
POPUPFIX /C: /D /F: /L: /R /S\r\n\n\
\t/C:\tcenter popup windows:  N A or W\n\
\t/D\trestore original plugin defaults\n\
\t/F:\treset focus when a dialog is closed:  N A or C\n\
\t/L:\tload plugin only if Console2 is detected:  Y or N  \n\
\t/R\trestore default options from the registry\n\
\t/S\tsave current options to the registry as default\n\
\n\
Any options may be combined except /D and /R.\n";


const wchar_t _ConEmuHelpText[]		= L"%_CONEMU\n\n\
Returns a nonzero value if TCC is running in a ConEmu tab,  \n\
or 0 if ConEmu is not detected.\n";


const wchar_t _Console2HelpText[]	= L"%_CONSOLE2\n\n\
Returns a nonzero value if TCC is running in a Console2 tab,  \n\
or 0 if Console2 was not detected.\n";


const wchar_t _TCmdTabHelpText[]	= L"%_TCMDTAB\n\n\
Returns a nonzero value if TCC is running in a Take Command tab,  \n\
or 0 if Take Command is not detected.\n";


const wchar_t _WinTermHelpText[]	= L"%_WINTERM\n\n\
Returns a nonzero value if TCC is running in Windows Terminal,  \n\
or 0 if Windows Terminal is not detected.\n";



int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_ARG_LEN] = L"%_batch";

	Qprintf( ERRHANDLE, L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables( shortbuf, 1 );
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s( shortbuf, MAX_ARG_LEN, L"%_batchname [%_batchline]  " );
	ExpandVariables( shortbuf, 1 );
	Qprintf( ERRHANDLE, shortbuf );

	return 0;
}



#include "ParseArgs.cpp"
#include "ForceFGWin.cpp"
#include "Ancestry.cpp"



BOOL GetVersion2( OSVERSIONINFOEXW *VerInfo )
{
	if ( !VerInfo )
		return -1;


	typedef NTSTATUS ( WINAPI _RtlGetVersionProc ) ( OSVERSIONINFOEXW *VerInfo );
	HMODULE NtDllHandle						= GetModuleHandle( L"NtDll.dll" );
	_RtlGetVersionProc *RtlGetVersionProc	= NULL;


	memset( VerInfo, 0x00, sizeof( OSVERSIONINFOEXW ) );
	VerInfo->dwOSVersionInfoSize = sizeof( OSVERSIONINFOEXW );

	if ( NtDllHandle )
		RtlGetVersionProc = (_RtlGetVersionProc*) GetProcAddress( NtDllHandle, "RtlGetVersion" );

	if ( RtlGetVersionProc ) {
		RtlGetVersionProc( VerInfo );
		if ( VerInfo->dwMajorVersion >= 6 )
			return TRUE;
	}

	memset( VerInfo, 0x00, sizeof( OSVERSIONINFOEXW ) );
	VerInfo->dwOSVersionInfoSize = sizeof( OSVERSIONINFOEXW );

	return ( GetVersionEx( (OSVERSIONINFOW*) VerInfo ) != 0 );
}



BOOL IsAncestor( DWORD Pid )
{
	if ( !Pid )
		return FALSE;

	
	if ( !MyAncestors || !NumAncestors )
		GetMyAncestors( &MyAncestors, &NumAncestors );

	if ( !MyAncestors || !NumAncestors )
		return FALSE;

	for ( unsigned int i = 0; i < NumAncestors; i++ )
		if ( Pid == MyAncestors[i] )
			return TRUE;

	return FALSE;
}



BOOL PipeShell()
{
	wchar_t temp[SHORT_BUF_LEN]			= L"%[_transient]%[_pipe]";


	ExpandVariables( temp, 1 );

	if ( wcscmp( temp, L"00" ) )
		return TRUE;
	else
		return FALSE;
}



int WindowInList( HWND hWnd )
{
	if ( !hWnd )
		return 0;

	for ( unsigned int i = 0; i < MAX_POPUP_WINDOWS; i++ )
		if ( hWnd == PopupWindowHandle[i] )
			return 1;

	return 0;
}



int AddWindowToList( HWND hWnd )
{
	if ( !hWnd )
		return 0;

	for ( unsigned int i = 0; i < MAX_POPUP_WINDOWS; i++ )
		if ( PopupWindowHandle[i] == NULL ) {
			PopupWindowHandle[i] = hWnd;
			NewPopupsFound = TRUE;
			return 0;
		}

	return 0;
}



int ForgetOldWindows()
{
	int rv = 0;

	for ( unsigned int i = 0; i < MAX_POPUP_WINDOWS; i++ )
		if ( PopupWindowHandle[i] )
			if ( !IsWindow( PopupWindowHandle[i] ) ) {
				PopupWindowHandle[i] = NULL;
				rv++;
			}

	return rv;
}



int WindowOffScreen( HWND hWnd )
{
	WINDOWINFO WindowInfo;

	if ( !GetWindowInfo( hWnd, &WindowInfo ) )
		return 0;

	if ( ( WindowInfo.rcClient.left < ScreenLeft ) || ( WindowInfo.rcClient.top < ScreenTop ) )
		return 1;

	if ( ( WindowInfo.rcClient.right > ScreenRight ) || ( WindowInfo.rcClient.bottom > ScreenBottom ) )
		return 1;

	return 0;
}



int CenterWindowOnMonitor( HWND hWnd )
{
	HMONITOR MonitorHandle = NULL;
	MONITORINFO MonitorInfo;
	RECT PopupFrame;
	int MonitorWidth = 0, MonitorHeight = 0, PopupWidth = 0, PopupHeight = 0;

	MonitorHandle = MonitorFromWindow( hWnd, MONITOR_DEFAULTTOPRIMARY );
	MonitorInfo.cbSize = sizeof( MONITORINFO );
	GetMonitorInfo( MonitorHandle, &MonitorInfo );

	GetWindowRect( hWnd, &PopupFrame );

	MonitorWidth	= MonitorInfo.rcWork.right	- MonitorInfo.rcWork.left;
	MonitorHeight	= MonitorInfo.rcWork.bottom	- MonitorInfo.rcWork.top;
	PopupWidth		= PopupFrame.right	- PopupFrame.left;
	PopupHeight		= PopupFrame.bottom	- PopupFrame.top;

	MoveWindow( hWnd, MonitorInfo.rcWork.left + ( ( MonitorWidth - PopupWidth ) / 2 ), MonitorInfo.rcWork.top + ( ( MonitorHeight - PopupHeight ) / 2 ), PopupWidth, PopupHeight, TRUE );
	return 0;
}



int CenterWindowOnWindow( HWND PopupWindow, HWND MainWindow )
{
	if ( !MainWindow )
		return CenterWindowOnMonitor( PopupWindow );
	else if ( WindowOffScreen( MainWindow ) )
		return CenterWindowOnMonitor( PopupWindow );
	else if ( !IsWindowVisible( MainWindow ) )
		return CenterWindowOnMonitor( PopupWindow );

	RECT MainFrame, PopupFrame;
	int MainWidth = 0, MainHeight = 0, PopupWidth = 0, PopupHeight = 0;

	GetWindowRect( MainWindow, &MainFrame );
	GetWindowRect( PopupWindow, &PopupFrame );

	MainWidth	= MainFrame.right - MainFrame.left;
	MainHeight	= MainFrame.bottom - MainFrame.top;
	PopupWidth	= PopupFrame.right - PopupFrame.left;
	PopupHeight	= PopupFrame.bottom - PopupFrame.top;

	MoveWindow( PopupWindow, MainFrame.left + ( ( MainWidth - PopupWidth ) / 2 ), MainFrame.top + ( ( MainHeight - PopupHeight ) / 2 ), PopupWidth, PopupHeight, TRUE );
	return 0;
}



int PutWindowNearWindow( HWND PopupWindow, HWND MainWindow )
{
	if ( !PopupWindow || !MainWindow )
		return -1;

	if ( !IsWindowVisible( MainWindow ) )
		return 1;

	RECT MainFrame, PopupFrame;
	int PopupWidth = 0, PopupHeight = 0;
	HMONITOR MonitorHandle = NULL;
	MONITORINFO MonitorInfo;
	int rc = 1;
	
	GetWindowRect( MainWindow, &MainFrame );
	GetWindowRect( PopupWindow, &PopupFrame );

	PopupWidth	= PopupFrame.right - PopupFrame.left;
	PopupHeight	= PopupFrame.bottom - PopupFrame.top;

	MonitorHandle = MonitorFromWindow( MainWindow, MONITOR_DEFAULTTONULL );
	if ( !MonitorHandle )
		return 1;

	MonitorInfo.cbSize = sizeof( MONITORINFO );
	GetMonitorInfo( MonitorHandle, &MonitorInfo );


	if ( ( ( MainFrame.top - PopupHeight ) >= MonitorInfo.rcWork.top ) && ( MainFrame.top < MonitorInfo.rcWork.bottom ) ) {
		if ( ( MainFrame.left >= MonitorInfo.rcWork.left ) && ( ( MainFrame.left + PopupWidth ) < MonitorInfo.rcWork.right ) ) {
			MoveWindow( PopupWindow, MainFrame.left, MainFrame.top - PopupHeight, PopupWidth, PopupHeight, TRUE );
			rc = 0;
		}
		else if ( ( ( MainFrame.right - PopupWidth ) >= MonitorInfo.rcWork.left ) && ( MainFrame.right < MonitorInfo.rcWork.right ) ) {
			MoveWindow( PopupWindow, MainFrame.right - PopupWidth, MainFrame.top - PopupHeight, PopupWidth, PopupHeight, TRUE );
			rc = 0;
		}
	}
	else if ( ( MainFrame.bottom >= MonitorInfo.rcWork.top ) && ( ( MainFrame.bottom + PopupHeight ) < MonitorInfo.rcWork.bottom ) ) {
		if ( ( MainFrame.left >= MonitorInfo.rcWork.left ) && ( ( MainFrame.left + PopupWidth ) < MonitorInfo.rcWork.right ) ) {
			MoveWindow( PopupWindow, MainFrame.left, MainFrame.bottom, PopupWidth, PopupHeight, TRUE );
			rc = 0;
		}
		else if ( ( ( MainFrame.right - PopupWidth ) >= MonitorInfo.rcWork.left ) && ( MainFrame.right < MonitorInfo.rcWork.right ) ) {
			MoveWindow( PopupWindow, MainFrame.right - PopupWidth, MainFrame.bottom, PopupWidth, PopupHeight, TRUE );
			rc = 0;
		}
	}
	else if ( ( ( MainFrame.left - PopupWidth ) >= MonitorInfo.rcWork.left ) && ( MainFrame.left < MonitorInfo.rcWork.right ) ) {
		if ( ( MainFrame.top >= MonitorInfo.rcWork.top ) && ( ( MainFrame.top + PopupHeight ) < MonitorInfo.rcWork.bottom ) ) {
			MoveWindow( PopupWindow, MainFrame.left - PopupWidth, MainFrame.top, PopupWidth, PopupHeight, TRUE );
			rc = 0;
		}
		else if ( ( ( MainFrame.bottom - PopupHeight ) >= MonitorInfo.rcWork.top ) && ( MainFrame.bottom < MonitorInfo.rcWork.bottom ) ) {
			MoveWindow( PopupWindow, MainFrame.left - PopupWidth, MainFrame.bottom - PopupHeight, PopupWidth, PopupHeight, TRUE );
			rc = 0;
		}
	}
	else if ( ( MainFrame.right >= MonitorInfo.rcWork.left ) && ( ( MainFrame.right + PopupWidth ) < MonitorInfo.rcWork.right ) ) {
		if ( ( MainFrame.top >= MonitorInfo.rcWork.top ) && ( ( MainFrame.top + PopupHeight ) < MonitorInfo.rcWork.bottom ) ) {
			MoveWindow( PopupWindow, MainFrame.right, MainFrame.top, PopupWidth, PopupHeight, TRUE );
			rc = 0;
		}
		else if ( ( ( MainFrame.bottom - PopupHeight ) >= MonitorInfo.rcWork.top ) && ( MainFrame.bottom < MonitorInfo.rcWork.bottom ) ) {
			MoveWindow( PopupWindow, MainFrame.right, MainFrame.bottom - PopupHeight, PopupWidth, PopupHeight, TRUE );
			rc = 0;
		}
	}

	return rc;
}



struct _FindWinByClassData {
	LPCTSTR ClassName;
	unsigned int MaxWindows;
	unsigned int n;
	HWND *hWnd;
	DWORD *Pid;
};



BOOL CALLBACK FindByClassProc$( HWND hWnd, LPARAM lParam )
{
	wchar_t Class[SHORT_BUF_LEN]			= L"";
	_FindWinByClassData *Data				= (_FindWinByClassData*) lParam;


	if ( !RealGetWindowClass( hWnd, Class, SHORT_BUF_LEN ) )
		return TRUE;

	if ( wcscmp( Class, Data->ClassName ) )
		return TRUE;

	if ( Data->n >= Data->MaxWindows )
		return FALSE;

	Data->hWnd[Data->n] = hWnd;

	if ( Data->Pid )
		GetWindowThreadProcessId( hWnd, Data->Pid + Data->n );

	Data->n += 1;
	return TRUE;
}



int FindWindowsByClass( LPCTSTR ClassName, HWND *hWnd, DWORD *Pid, unsigned int MaxWindows )
{
	if ( !ClassName || !hWnd || !MaxWindows )
		return -1;
	else if ( ClassName[0] == '\0' )
		return -1;


	_FindWinByClassData Data;


	memset( hWnd, 0x00, MaxWindows * sizeof( HWND ) );
	if ( Pid )
		memset( Pid, 0x00, MaxWindows * sizeof( DWORD ) );

	Data.ClassName	= ClassName;
	Data.MaxWindows	= MaxWindows;
	Data.n			= 0;
	Data.hWnd		= hWnd;
	Data.Pid		= Pid;


	EnumWindows( FindByClassProc$, (LPARAM) &Data );
	return Data.n;
}



HWND MainWindowHandle()
{
	if ( Console2WindowHandle )
		return Console2WindowHandle;

	if ( TCmdWindowHandle )
		return TCmdWindowHandle;

	if ( ConEmuWindowHandle )
		return ConEmuWindowHandle;

	if ( WinTermWindowHandle )
		return WinTermWindowHandle;

	return ConsoleWindowHandle;
}



int GetScreenSize()
{
	ScreenLeft		= GetSystemMetrics( SM_XVIRTUALSCREEN );
	ScreenTop		= GetSystemMetrics( SM_YVIRTUALSCREEN );
	ScreenWidth		= GetSystemMetrics( SM_CXVIRTUALSCREEN );
	ScreenHeight	= GetSystemMetrics( SM_CYVIRTUALSCREEN );

	ScreenRight		= ScreenLeft + ScreenWidth;
	ScreenBottom	= ScreenTop + ScreenHeight;

	//  wprintf( L"   * Left = %d   Top = %d   Right = %d   Bottom = %d\n", ScreenLeft, ScreenTop, ScreenRight, ScreenBottom );
	return 0;
}




int DummyConsoleOutput( unsigned int n )
{
	for ( unsigned int i = 0; i < n; i++ )
		_cputws( L"A\bB\bC\bD\b \b" );

	return 0;
}



BOOL CALLBACK TakeCmdWindowCheck( HWND hWnd, LPARAM lParam )
{
	if ( !lParam )
		return TRUE;


	wchar_t ClassName[SHORT_BUF_LEN] = L"";

	if ( !RealGetWindowClass( hWnd, ClassName, SHORT_BUF_LEN ) )
		return TRUE;

	if ( _wcsnicmp( ClassName, TCmdClass, TCmdClassLen ) )
		if ( _wcsicmp( ClassName, TCmdClass2 ) )
			return TRUE;

	HWND *TCmdHWnd = (HWND*) lParam;

	for ( int i = 0; i < MAX_CONSOLE2_WINDOWS; i++ ) {
		if ( TCmdHWnd[i] == NULL ) {
			TCmdHWnd[i] = hWnd;
			return TRUE;
		}
	}

	return TRUE;
}



HWND GetTakeCommandWindow()
{
	HWND TCmdHWnd[MAX_CONSOLE2_WINDOWS];
	DWORD TCmdPid[MAX_CONSOLE2_WINDOWS];

	for ( int i = 0; i < MAX_CONSOLE2_WINDOWS; i++ )
		TCmdHWnd[i] = NULL, TCmdPid[i] = 0;


	EnumWindows( TakeCmdWindowCheck, (LPARAM) TCmdHWnd );

	if ( TCmdHWnd[0] == NULL )
		return NULL;

	for ( int i = 0; i < MAX_CONSOLE2_WINDOWS; i++ )
		if ( TCmdHWnd[i] )
			GetWindowThreadProcessId( TCmdHWnd[i], TCmdPid + i );

	/*
	wprintf( L"\n   * Take Command windows:\n\n" );
	for ( int i = 0; i < MAX_CONSOLE2_WINDOWS; i++ )
		if ( TCmdHWnd[i] )
			wprintf( L"   * %02i:  hWnd = 0x%08x   Pid = 0x%08x\n", i, TCmdHWnd[i], TCmdPid[i] );
	*/

	DWORD BufferSize			= 64;
	DWORD ProcessCount			= BufferSize + 1;
	DWORD *ProcessListBuffer	= NULL;


	DummyConsoleOutput( 256 );

	while ( ProcessCount > BufferSize ) {

		if ( ProcessListBuffer ) {
			free( ProcessListBuffer );
			ProcessListBuffer = NULL;
		}

		ProcessListBuffer = (DWORD*) malloc( BufferSize * sizeof( DWORD ) );

		if ( !ProcessListBuffer )
			return NULL;

		ProcessCount = GetConsoleProcessList( ProcessListBuffer, BufferSize );
	}

	if ( ProcessCount == 0 ) {
		if ( ProcessListBuffer )
			free( ProcessListBuffer );

		return NULL;
	}

	/*
	wprintf( L"\n   * Attached processes:\n\n" );
	for ( unsigned int i = 0; i < ProcessCount; i++ )
		if ( ProcessListBuffer[i] )
			wprintf( L"   * %02u:  0x%08x\n", i, ProcessListBuffer[i] );

	wprintf( L"\n" );
	*/

	for ( unsigned int i = 0; i < ProcessCount; i++ ) {
		if ( ProcessListBuffer[i] ) {
			for ( int j = 0; j < MAX_CONSOLE2_WINDOWS; j++ )
				if ( ProcessListBuffer[i] == TCmdPid[j] )
					return TCmdHWnd[j];
		}
	}

	return NULL;
}



int ParseInt( LPTSTR inString, unsigned int *outValue, wchar_t **EndChar )
{
	if ( !inString || !outValue )
		return -1;

	unsigned int value = 0;
	wchar_t *sep1 = NULL;

	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) ) {
		//  parse hexadecimal

		value = wcstoul( inString + 2, &sep1, 16 );

		if ( errno )
			return -1;
		else if ( !EndChar && *sep1 )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}
	else if ( iswdigit( inString[0] ) ) {
		//  parse unsigned

		value = wcstoul( inString, &sep1, 10 );

		if ( errno )
			return -1;
		else if ( !EndChar && *sep1 )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}

	return -1;
}



LPTSTR _FilenamePart( LPTSTR Filename, LPTSTR outFilename )
{
	if ( !Filename )
		return NULL;
	else if ( Filename[0] == '\0' )
		return NULL;


	BOOL FoundSlash			= FALSE;
	size_t EndOfPathPart	= -1;


	for ( size_t i = 0; Filename[i]; i++ ) {
		if ( ( Filename[i] == '/' ) || ( Filename[i] == '\\' ) ) {
			EndOfPathPart = i;
			FoundSlash = TRUE;
		}
		else if ( ( Filename[i] == ':' ) && !FoundSlash )
			EndOfPathPart = i;
	}

	if ( EndOfPathPart == -1 ) {
		if ( outFilename ) {
			wcscpy_s( outFilename, MAX_FILENAME_LEN, Filename );
			return outFilename;
		}
		else
			return Filename;
	}

	if ( outFilename ) {
		wcscpy_s( outFilename, MAX_FILENAME_LEN, Filename + EndOfPathPart + 1 );
		return outFilename;
	}
	else
		return Filename + EndOfPathPart + 1;
}



int GetConEmuCFilename( LPTSTR outString )
{
	if ( !outString )
		return -1;


	DWORD BufferSize			= 64;
	DWORD ProcessCount			= BufferSize + 1;
	DWORD *ProcessListBuffer	= NULL;


	while ( ProcessCount > BufferSize ) {

		if ( ProcessListBuffer ) {
			free( ProcessListBuffer );
			ProcessListBuffer = NULL;
		}

		ProcessListBuffer = (DWORD*) malloc( BufferSize * sizeof( DWORD ) );

		if ( !ProcessListBuffer )
			return 1;

		ProcessCount = GetConsoleProcessList( ProcessListBuffer, BufferSize );
	}

	if ( ProcessCount == 0 ) {
		if ( ProcessListBuffer )
			free( ProcessListBuffer );

		return 1;
	}


	wchar_t Filename[MAX_FILENAME_LEN]			= L"";
	LPTSTR BaseName			= NULL;
	HANDLE ProcessHandle	= NULL;
	int rv					= 1;


	for ( unsigned int i = 0; i < ProcessCount; i++ ) {
		Filename[0] = '\0';
		
		ProcessHandle = OpenProcess( PROCESS_QUERY_INFORMATION | PROCESS_VM_READ, FALSE, ProcessListBuffer[i] );

		if ( ProcessHandle ) {
			GetModuleFileNameEx( ProcessHandle, NULL, Filename, MAX_FILENAME_LEN );
			CloseHandle( ProcessHandle );
			ProcessHandle = NULL;
		}

		if ( Filename[0] ) {
			BaseName = _FilenamePart( Filename, NULL );

			if ( BaseName ) {
				if ( !_wcsicmp( BaseName, L"ConEmuC64.exe" ) ) {
					wcscpy_s( outString, MAX_FILENAME_LEN, Filename );
					rv = 0;
					break;
				}
				else if ( !_wcsicmp( BaseName, L"ConEmuC.exe" ) ) {
					wcscpy_s( outString, MAX_FILENAME_LEN, Filename );
					rv = 0;
					break;
				}
			}
		}
	}

	free( ProcessListBuffer );
	return rv;
}



BOOL CALLBACK ConEmuWindowCheck( HWND hWnd, LPARAM lParam )
{
	if ( !lParam )
		return TRUE;


	wchar_t ClassName[SHORT_BUF_LEN] = L"";

	if ( !RealGetWindowClass( hWnd, ClassName, SHORT_BUF_LEN ) )
		return TRUE;

	if ( wcscmp( ClassName, ConEmuClass ) )
		return TRUE;

	HWND *ConEmuWindow = (HWND*) lParam;

	if ( *ConEmuWindow == NULL )
		*ConEmuWindow = hWnd;

	return TRUE;
}



HWND GetConEmuWindow()
{
	wchar_t Filename[MAX_FILENAME_LEN]			= L"";
	wchar_t CmdBuffer[MAX_FILENAME_LEN + 32]	= L"";
	HWND ConEmuHandle							= NULL;
	unsigned int Value							= 0;


	if ( GetConEmuCFilename( Filename ) )
		return NULL;

	wsprintf( CmdBuffer, L"\"%s\" /GUIMACRO IsConEmu > nul:", Filename );

	if ( Command( CmdBuffer, 0 ) )
		return NULL;


	Filename[0] = '\0';
	GetEnvironmentVariable( L"ConEmuHWND", Filename, MAX_FILENAME_LEN );
	if ( Filename[0] ) {
		if ( !ParseInt( Filename, &Value, NULL ) ) {
			Filename[0] = '\0';
			RealGetWindowClass( (HWND) Value, Filename, MAX_FILENAME_LEN );

			if ( !wcscmp( Filename, ConEmuClass ) )
				return (HWND) Value;
		}
	}

	EnumWindows( ConEmuWindowCheck, (LPARAM) &ConEmuHandle );

	return ConEmuHandle;
}



HWND GetWindowsTerminalWindow()
{
	wchar_t temp[SHORT_BUF_LEN]			= L"";

	HWND hWnd[MAX_WINTERM_WINDOWS];
	DWORD Pid[MAX_WINTERM_WINDOWS];

	int NumFound					= 0;


	RealGetWindowClass( GetConsoleWindow(), temp, SHORT_BUF_LEN );
	if ( _wcsicmp( temp, L"PseudoConsoleWindow" ) )
		return 0;

	NumFound = FindWindowsByClass( L"CASCADIA_HOSTING_WINDOW_CLASS", hWnd, Pid, MAX_WINTERM_WINDOWS );
	if ( NumFound < 1 )
		return NULL;

	for ( int i = 0; i < NumFound; i++ )
		if ( IsAncestor( Pid[i] ) )
			return hWnd[i];

	return NULL;
}



int MoveHiddenConsoleWindow( HWND Window )
{
	HWND ConsoleWindow		= GetConsoleWindow();
	RECT WinTermFrame		= { 0, 0, 0, 0 };
	RECT ConsoleFrame		= { 0, 0, 0, 0 };

	int rv					= 0;


	if ( GetWindowRect( Window, &WinTermFrame ) && GetWindowRect( ConsoleWindow, &ConsoleFrame ) )
		if ( memcmp( &WinTermFrame, &ConsoleFrame, sizeof( RECT ) ) )
			if ( SetWindowPos( ConsoleWindow, HWND_TOP, WinTermFrame.left, WinTermFrame.top, WinTermFrame.right - WinTermFrame.left, WinTermFrame.bottom - WinTermFrame.top, SWP_HIDEWINDOW | SWP_NOACTIVATE | SWP_NOZORDER ) )
				rv = 1;

	return rv;
}



BOOL CALLBACK CheckWindow( HWND hWnd, LPARAM lParam )
{
	if ( !IsWindowVisible( hWnd ) )
		return TRUE;

	wchar_t ClassName[SHORT_BUF_LEN] = L"";

	RealGetWindowClass( hWnd, ClassName, SHORT_BUF_LEN );

	if ( wcscmp( ClassName, PopupClass ) )
		if ( wcscmp( ClassName, DialogClass ) )
			return TRUE;

	if ( WindowInList( hWnd ) )
		return TRUE;

	AddWindowToList( hWnd );

	ForceForegroundWindow( hWnd );

	if ( CenterPopups == CENTER_REPOSITION )
		PutWindowNearWindow( hWnd, MainWindowHandle() );

	if ( WindowOffScreen( hWnd ) || ( CenterPopups == CENTER_ALWAYS ) )
		CenterWindowOnWindow( hWnd, MainWindowHandle() );

	return TRUE;
}



VOID CALLBACK MyTimerProc( HWND hWnd, UINT uMsg, UINT_PTR idEvent, DWORD dwTime )
{
	HWND MainWindow = MainWindowHandle();

	int visible = IsWindowVisible( ConsoleWindowHandle );


	if ( visible != ConsoleWindowVisible ) {
		ConsoleWindowVisible = visible;

		if ( !Console2WindowHandle && !WinTermWindowHandle ) {
			TCmdWindowHandle = GetTakeCommandWindow();

			if ( TCmdWindowHandle )
				ConEmuWindowHandle = NULL;
			else
				ConEmuWindowHandle = GetConEmuWindow();
		}
	}

	if ( Console2WindowHandle )
		MoveHiddenConsoleWindow( Console2WindowHandle );
	else if ( WinTermWindowHandle ) {
		WinTermWindowHandle = GetWindowsTerminalWindow();
		MoveHiddenConsoleWindow( WinTermWindowHandle );
	}


	if ( RefocusFlag ) {
		if ( MainWindow )
			if ( GetForegroundWindow() != MainWindow )
				ForceForegroundWindow( MainWindow );

		RefocusFlag = FALSE;
		return;
	}

	if ( ForgetOldWindows() ) {
		if ( RestoreFocus == RESTORE_FOCUS_CONSOLE2 )
			RefocusFlag = ( WinTermWindowHandle || Console2WindowHandle );
		else if ( RestoreFocus )
			RefocusFlag = TRUE;
	}

	NewPopupsFound = FALSE;
	EnumThreadWindows( MyThread, CheckWindow, 0 );

	if ( NewPopupsFound && RefocusFlag )
		RefocusFlag = FALSE;

	ResCheckCount = ( ResCheckCount + 1 ) % ResCheckCycles;
	if ( ResCheckCount == 0 )
		GetScreenSize();

	return;
}



unsigned int TCVerMajor()
{
	wchar_t Buffer[SHORT_BUF_LEN] = L"%_4ver";
	ExpandVariables( Buffer, 1 );
	
	return wcstoul( Buffer, NULL, 10 );
}



int GetEnvironmentValue( LPCTSTR VarName, int Minimum, int Maximum, int *outValue )
{
	if ( !VarName || !outValue )
		return -1;

	wchar_t Buffer[SHORT_BUF_LEN] = L"";
	wchar_t *ch = Buffer;
	int Value = 0, Radix = 10;
	

	if ( !GetEnvironmentVariable( VarName, Buffer, SHORT_BUF_LEN ) )
		return 1;

	while ( iswspace( *ch ) )
		ch++;

	if ( ( *ch  == '0' ) && ( towupper( *(ch + 1) ) == 'X' ) ) {
		ch += 2;
		Radix = 16;
	}

	_set_errno( 0 );
	Value = wcstol( ch, &ch, Radix );

	if ( errno )
		return 1;

	while ( iswupper( *ch ) )
		ch++;

	if ( *ch )
		return 1;

	if ( ( Value < Minimum ) || ( Value > Maximum ) )
		return 2;

	*outValue = Value;
	return 0;
}



int GetRegistryValue( HKEY RootKey, LPCTSTR RegKeyName, LPCTSTR RegEntryName, int Minimum, int Maximum, int *outValue )
{
	if ( !RegKeyName || !RegEntryName || !outValue )
		return -1;


	HKEY KeyHandle = NULL;
	DWORD Data = 0, Type = 0, Size = sizeof( DWORD );
	int rv = 0;

	if ( RegOpenKeyEx( RootKey, RegKeyName, 0, KEY_READ, &KeyHandle ) != ERROR_SUCCESS )
		return -1;

	rv = RegQueryValueEx( KeyHandle, RegEntryName, NULL, &Type, (BYTE*) &Data, &Size );

	RegCloseKey( KeyHandle );

	if ( ( rv == ERROR_SUCCESS ) && ( Type == REG_DWORD ) ) {
		if ( ( ( (int) Data ) < Minimum ) || ( (int) ( Data ) > Maximum ) )
			return 2;

		*outValue = Data;
		return 0;
	}

	return 1;
}



int GetDefaultValue( LPCTSTR VarName, LPCTSTR RegKeyName, LPCTSTR RegEntryName, int Minimum, int Maximum, int Default )
{
	int Value = 0;

	if ( !GetEnvironmentValue( VarName, Minimum, Maximum, &Value ) )
		return Value;

	if ( !GetRegistryValue( HKEY_CURRENT_USER, RegKeyName, RegEntryName, Minimum, Maximum, &Value ) )
		return Value;

	if ( !GetRegistryValue( HKEY_LOCAL_MACHINE, RegKeyName, RegEntryName, Minimum, Maximum, &Value ) )
		return Value;

	return Default;
}



int GetOptionValue( LPTSTR Prefix, LPTSTR Suffix, int *outValue )
{
	if ( !Prefix || !Suffix || !outValue )
		return -1;

	wchar_t Buffer[SHORT_BUF_LEN] = L"";
	int Value = 0;
	wchar_t *sep1 = NULL;


	wsprintf( Buffer, L"%%@option[%s%s]", Prefix, Suffix );
	ExpandVariables( Buffer, 1 );

	_set_errno( 0 );
	Value = wcstol( Buffer, &sep1, 10 );

	if ( errno || *sep1 )
		return 1;

	*outValue = Value;
	return 0;
}



BOOL CALLBACK Console2WindowCheck( HWND hWnd, LPARAM lParam )
{
	if ( !lParam )
		return TRUE;


	wchar_t ClassName[SHORT_BUF_LEN] = L"";

	if ( !RealGetWindowClass( hWnd, ClassName, SHORT_BUF_LEN ) )
		return TRUE;

	if ( wcscmp( ClassName, Console2Class ) )
		return TRUE;

	HWND *Console2hWnd = (HWND*) lParam;

	for ( int i = 0; i < MAX_CONSOLE2_WINDOWS; i++ ) {
		if ( Console2hWnd[i] == NULL ) {
			Console2hWnd[i] = hWnd;
			return TRUE;
		}
	}

	return TRUE;
}



int SameConsoleWindow( DWORD Pid )
{
	DWORD ConsoleWindowPid = 0;

	GetWindowThreadProcessId( GetConsoleWindow(), &ConsoleWindowPid );

	if ( ConsoleWindowPid == Pid )
		return 1;
	else
		return 0;
}



HWND Console2Window()
{
	HWND Console2hWnd[MAX_CONSOLE2_WINDOWS];
	DWORD Console2Pid[MAX_CONSOLE2_WINDOWS];

	for ( int i = 0; i < MAX_CONSOLE2_WINDOWS; i++ )
		Console2hWnd[i] = NULL, Console2Pid[i] = 0;


	EnumWindows( Console2WindowCheck, (LPARAM) Console2hWnd );

	if ( Console2hWnd[0] == NULL )
		return NULL;

	for ( int i = 0; i < MAX_CONSOLE2_WINDOWS; i++ )
		if ( Console2hWnd[i] )
			GetWindowThreadProcessId( Console2hWnd[i], Console2Pid + i );


	DWORD MyPid = GetCurrentProcessId();
	DWORD Pid = MyPid, ParentPid = NULL, LastPid = NULL;

	do {
		ParentPid = GetParentPID( Pid );
		if ( ( ParentPid == Pid ) || ( ParentPid == 0xffffffff ) )
			break;

		LastPid = Pid;
		Pid = ParentPid;

		for ( int i = 0; i < MAX_CONSOLE2_WINDOWS; i++ ) {
			if ( Pid == Console2Pid[i] ) {
				if ( LastPid == MyPid )
					return Console2hWnd[i];
				else if ( SameConsoleWindow( LastPid ) )
					return Console2hWnd[i];
				else
					return NULL;
			}
		}

	} while ( TRUE );

	return NULL;
}



BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}



DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/popupfix.html";
	piInfo.pszDescription	= (LPTSTR) L"Ensures that TCC\u2019s popup dialog boxes are visible.";
	piInfo.pszFunctions		= (LPTSTR) Features;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

	return &piInfo;
}



DLLExports BOOL WINAPI InitializePlugin( void )
{
	wchar_t shortbuf[SHORT_BUF_LEN] = L"%[_transient]%[_pipe]%[noloadmsg]";
	ExpandVariables( shortbuf, 1 );


	unsigned int TCCVersion = TCVerMajor();

	if ( TCCVersion < 9 ) {
		if ( !PipeShell() ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"This plugin is for Take Command v9 or later!\r\nPlugin not loaded.\r\n" );
		}

		return -1;
	}

	MyThread = GetCurrentThreadId();
	GetVersion2( &OSVerInfo );
	GetMyAncestors( &MyAncestors, &NumAncestors );

	for ( int i = 0; i < MAX_POPUP_WINDOWS; i++ )
		PopupWindowHandle[i] = NULL;

	GetScreenSize();
	ResCheckCount = 0;

	TimerMS = GetDefaultValue( TimerMSEnvName, PluginKeyName, TimerMSRegName, TIMER_MS_MIN, TIMER_MS_MAX, TIMER_MS_DEFAULT );
	ResCheckCycles = ( RES_CHECK_SECONDS * 1000 ) / TimerMS;

	//  wprintf( L"   * TimerMS = %d   ResCheckCycles = %d\n", TimerMS, ResCheckCycles );

	ConsoleWindowHandle		= GetConsoleWindow();
	ConsoleWindowVisible	= IsWindowVisible( ConsoleWindowHandle );

	Console2WindowHandle	= Console2Window();
	WinTermWindowHandle		= GetWindowsTerminalWindow();

	if ( !Console2WindowHandle && !WinTermWindowHandle ) {
		TCmdWindowHandle = GetTakeCommandWindow();

		if ( !TCmdWindowHandle )
			ConEmuWindowHandle = GetConEmuWindow();
	}
	

	Console2Only = GetDefaultValue( Console2OnlyEnvName, PluginKeyName, Console2OnlyRegName, INT_MIN, INT_MAX, FALSE );

	if ( Console2Only && !Console2WindowHandle && !WinTermWindowHandle ) {
		if ( !wcscmp( shortbuf, L"00" ) ) {
			Printf( L"%s: Neither Console2 nor Windows Terminal was detected.\r\n" );
			Printf( L"    Plugin not loaded.\r\n", PLUGIN_NAME );
		}

		return -1;
	}

	CenterPopups		= GetDefaultValue( AlwaysCenterEnvName, PluginKeyName, AlwaysCenterRegName, INT_MIN, INT_MAX, CENTER_ALWAYS );
	RestoreFocus		= GetDefaultValue( RestoreFocusEnvName, PluginKeyName, RestoreFocusRegName, INT_MIN, INT_MAX, RESTORE_FOCUS_CONSOLE2 );


	TimerID = SetTimer( NULL, 0, TimerMS, MyTimerProc );

	if ( TimerID == 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error creating timer!  Plugin not loaded.\r\n" );
		return -1;
	}


	if ( Console2WindowHandle )
		MoveHiddenConsoleWindow( Console2WindowHandle );
	else if ( WinTermWindowHandle )
		MoveHiddenConsoleWindow( WinTermWindowHandle );

	if ( !wcscmp( shortbuf, L"00" ) )
		Printf( L"%s v%d.%d.%d loaded.\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );

	return 0;
}



DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{
	KillTimer( NULL, TimerID );
	free( MyAncestors );

	return 0;
}



int ShowCmdHelp( LPCTSTR HelpText, BOOL ShowVersion )
{
	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion )
		Printf( L"\r\n\t\t\t\t%s  %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );

	return 0;
}



int TestYorN( LPTSTR inString, int *outValue )
{
	if ( !inString || !outValue )
		return -1;

	int Value = -1;

	if ( !_wcsicmp( inString, L"Yes" ) )
		Value = 1;
	else if ( !_wcsicmp( inString, L"No" ) )
		Value = 0;
	else if ( !_wcsicmp( inString, L"Y" ) )
		Value = 1;
	else if ( !_wcsicmp( inString, L"N" ) )
		Value = 0;
	else if ( !_wcsicmp( inString, L"On" ) )
		Value = 1;
	else if ( !_wcsicmp( inString, L"Off" ) )
		Value = 0;

	if ( Value == -1 ) {
		ParseInt( inString, (unsigned int*) &Value, NULL );

		if ( ( Value != 0 ) && ( Value != 1 ) )
			Value = -1;
	}

	if ( Value == -1 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Should be YES or NO: \"%s\"\r\n", inString );
		return 1;
	}

	*outValue = Value;
	return 0;
}



int IsColonOrEquals( wchar_t ch )
{
	return ( ( ch == ':' ) || ( ch == '=' ) );
}



/*
DLLExports int WINAPI conproc( LPTSTR inString )
{
	const int MaxProcesses = 256;
	int NumProcesses = 0;

	DWORD Buffer[MaxProcesses];
	
	wchar_t Filename[MAX_FILENAME_LEN]	= L"";
	HANDLE ProcessHandle				= NULL;


	NumProcesses = GetConsoleProcessList( Buffer, MaxProcesses );

	wprintf( L"   * NumProcesses = %d\n", NumProcesses );

	if ( ( NumProcesses ) && ( NumProcesses <= MaxProcesses ) )
		for ( int i = 0; i < NumProcesses; i++ ) {
			Filename[0] = '\0';

			ProcessHandle = OpenProcess( PROCESS_QUERY_INFORMATION | PROCESS_VM_READ, FALSE, Buffer[i] );

			if ( ProcessHandle ) {
				GetModuleFileNameEx( ProcessHandle, NULL, Filename, MAX_FILENAME_LEN );
				CloseHandle( ProcessHandle );
				ProcessHandle = NULL;
			}

			wprintf( L"   * %3d   0x%08x   %s\n", i, Buffer[i], Filename );
		}
	
	return 0;
}
*/



DLLExports int WINAPI _console2( LPTSTR outString )
{
	if ( !outString )
		return 1;


	wcscpy_s( outString, MAX_ARG_LEN, L"0" );

	if ( Console2WindowHandle )
		wsprintf( outString, L"0x%08x", Console2WindowHandle );

	return 0;
}



DLLExports int WINAPI _conemu( LPTSTR outString )
{
	if ( !outString )
		return 1;


	HWND ConEmuHandle = GetConEmuWindow();

	if ( ConEmuHandle )
		wsprintf( outString, L"0x%08x", ConEmuHandle );
	else
		wcscpy_s( outString, MAX_ARG_LEN, L"0" );

	return 0;
}



DLLExports int WINAPI _tcmdtab( LPTSTR outString )
{
	if ( !outString )
		return 1;


	wcscpy_s( outString, MAX_ARG_LEN, L"0" );

	HWND TakeCommandWindow = NULL;

	TakeCommandWindow = GetTakeCommandWindow();

	if ( TakeCommandWindow )
		wsprintf( outString, L"0x%08x", TakeCommandWindow );

	return 0;
}



DLLExports int WINAPI _winterm( LPTSTR outString )
{
	if ( !outString )
		return 1;


	wcscpy_s( outString, MAX_ARG_LEN, L"0" );

	if ( WinTermWindowHandle )
		wsprintf( outString, L"0x%08X", WinTermWindowHandle );

	return 0;
}



int ShowYesNo( int Value )
{
	if ( Value )
		QPuts( L"Yes" );
	else
		QPuts( L"No" );

	return 0;
}



DLLExports int WINAPI popupfix( LPTSTR inString )
{
	if ( !inString )
		return -1;


	BOOL Save = FALSE, Restore = FALSE, PluginDefaults = FALSE;
	int NewTimerMS				= -1;
	int NewCenterPopups			= -1;
	int NewRestoreFocus			= -1;
	int NewConsole2Only			= -1;

	int rv = 0;

	
	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_FORCE_UPPERCASE );

	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcsncmp( arg[i], L"/?", 2 ) )
				return ShowCmdHelp( PopupFixHelpText, ( arg[i][2] == '?' ) );
			else if ( !wcsncmp( arg[i], L"/C", 2 ) ) {
				wchar_t *ch = arg[i] +2;
				if ( IsColonOrEquals( *ch ) )
					ch++;

				if ( *ch == 'N' )
					NewCenterPopups = CENTER_AS_NEEDED;
				else if ( ( *ch == 'A' ) || ( *ch == 'Y' ) )
					NewCenterPopups = CENTER_ALWAYS;
				else if ( *ch == 'W' )
					NewCenterPopups = CENTER_REPOSITION;
				else {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Should be N, A, or W: \"%s\"\r\n", arg[i] );
					return 1;
				}
			}
			else if ( !wcscmp( arg[i], L"/D" ) )
				PluginDefaults = TRUE;
			else if ( !wcsncmp( arg[i], L"/F", 2 ) ) {
				wchar_t *ch = arg[i] +2;
				if ( IsColonOrEquals( *ch ) )
					ch++;

				NewRestoreFocus = -1;

				if ( *ch == 'N' )
					NewRestoreFocus = RESTORE_FOCUS_NEVER;
				else if ( *ch == 'A' )
					NewRestoreFocus = RESTORE_FOCUS_ALWAYS;
				else if ( *ch == 'C' )
					NewRestoreFocus = RESTORE_FOCUS_CONSOLE2;
				else {
					ParseInt( ch, (unsigned int*) &NewRestoreFocus, NULL );

					if ( ( NewRestoreFocus < 0 ) || ( NewRestoreFocus > RESTORE_FOCUS_MAX_VALUE ) ) {
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Should be NEVER, ALWAYS, or CONSOLE2: \"%s\"\r\n", ch );
						return 1;
					}
				}
			}
			else if ( !wcsncmp( arg[i], L"/L", 2 ) ) {
				wchar_t *ch = arg[i] +2;
				if ( IsColonOrEquals( *ch ) )
					ch++;

				if ( TestYorN( ch, &NewConsole2Only ) )
					return 1;
			}
			else if ( !wcscmp( arg[i], L"/R" ) )
				Restore = TRUE;
			else if ( !wcscmp( arg[i], L"/S" ) )
				Save = TRUE;
			else if ( !wcsncmp( arg[i], L"/T", 2 ) ) {
				wchar_t *ch = arg[i] + 2;
				if ( IsColonOrEquals( *ch ) )
					ch++;

				if ( ParseInt( ch, (unsigned int*) &NewTimerMS, NULL ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad value: \"%s\"\r\n", ch );
					return 1;
				}
				else if ( ( NewTimerMS < TIMER_MS_MIN ) || ( NewTimerMS > TIMER_MS_MAX ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"TimerMS must be %u - %u\r\n", TIMER_MS_MIN, TIMER_MS_MAX );
					return 1;
				}
			}
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}
			else {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
				return 1;
			}
		}
	}


	if ( PluginDefaults && Restore ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Can\u2019t combine /D and /R\r\n" );
		return 1;
	}

	if ( Restore ) {
		TimerMS				= GetDefaultValue( NULL, PluginKeyName, TimerMSRegName, TIMER_MS_MIN, TIMER_MS_MAX, TIMER_MS_DEFAULT );
		Console2Only		= GetDefaultValue( NULL, PluginKeyName, Console2OnlyRegName, INT_MIN, INT_MAX, FALSE );
		CenterPopups		= GetDefaultValue( NULL, PluginKeyName, AlwaysCenterRegName, INT_MIN, INT_MAX, CENTER_ALWAYS );
		RestoreFocus		= GetDefaultValue( NULL, PluginKeyName, RestoreFocusRegName, INT_MIN, INT_MAX, RESTORE_FOCUS_CONSOLE2 );
	}
	else if ( PluginDefaults ) {
		TimerMS				= TIMER_MS_DEFAULT;
		Console2Only		= FALSE;
		CenterPopups		= CENTER_ALWAYS;
		RestoreFocus		= RESTORE_FOCUS_CONSOLE2;
	}


	if ( NewTimerMS != -1 )
		TimerMS = NewTimerMS;

	if ( NewCenterPopups != -1 )
		CenterPopups = NewCenterPopups;

	if ( NewRestoreFocus != -1 )
		RestoreFocus = NewRestoreFocus;

	if ( NewConsole2Only != -1 )
		Console2Only = NewConsole2Only;


	Printf( L"\r\nPopupFix plugin v%u.%u.%u options:\r\n", VER_MAJOR, VER_MINOR, VER_BUILD );

	QPuts( L"\r\n   Center popup windows .................  /C : " );
	
	switch ( CenterPopups ) {
		case CENTER_ALWAYS :
			QPuts( L"Always" );
			break;

		case CENTER_REPOSITION :
			QPuts( L"put beside Window" );
			break;

		default :
			QPuts( L"as Needed" );
	}

	QPuts( L"\r\n   Reset focus after a dialog closes ....  /F : " );
	
	switch ( RestoreFocus ) {
		case RESTORE_FOCUS_NEVER :
			QPuts( L"Never" );
			break;

		case RESTORE_FOCUS_CONSOLE2 :
			QPuts( L"Console2 and WinTerm only" );
			break;

		default :
			QPuts( L"Always" );
	}

	QPuts( L"\r\n   Load only in Console2 / WinTerm ......  /L : " );
	ShowYesNo( Console2Only );

	if ( ( NewTimerMS != -1 ) || ( TimerMS != TIMER_MS_DEFAULT ) )
		Printf( L"\r\n   Timer rate ...........................  /T : %d ms\r\n\r\n", TimerMS );

	QPuts( L"\r\n" );


	if ( Save ) {
		HKEY KeyHandle = NULL;
		
		if ( RegCreateKeyEx( HKEY_CURRENT_USER, PluginKeyName, 0, NULL, REG_OPTION_NON_VOLATILE, KEY_READ | KEY_WRITE, NULL, &KeyHandle, NULL ) == ERROR_SUCCESS ) {

			if ( RegSetValueEx( KeyHandle, TimerMSRegName, 0, REG_DWORD, (BYTE*) &TimerMS, sizeof( DWORD ) ) != ERROR_SUCCESS )
				rv = 2;

			if ( RegSetValueEx( KeyHandle, Console2OnlyRegName, 0, REG_DWORD, (BYTE*) &Console2Only, sizeof( DWORD ) ) != ERROR_SUCCESS )
				rv = 2;

			if ( RegSetValueEx( KeyHandle, AlwaysCenterRegName, 0, REG_DWORD, (BYTE*) &CenterPopups, sizeof( DWORD ) ) != ERROR_SUCCESS )
				rv = 2;

			if ( RegSetValueEx( KeyHandle, RestoreFocusRegName, 0, REG_DWORD, (BYTE*) &RestoreFocus, sizeof( DWORD ) ) != ERROR_SUCCESS )
				rv = 2;

			RegCloseKey( KeyHandle );
		}
		else
			rv = 2;

		if ( rv == 0 )
			QPuts( L"\r\nOptions saved to the registry.\r\n" );
		else {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error saving options to the registry\r\n" );
		}
	}

	return rv;
}



#ifndef	NEWHELP_VERSION

LPTSTR UsageText( const wchar_t *inString )
{
	if ( !inString )
		return NULL;


	wchar_t *ch = wcschr( (wchar_t*) inString, '\n' );


	if ( !ch )
		return NULL;

	while ( *ch == '\n' )
		ch++;

	return ch;
}

#endif	//	NEWHELP_VERSION



DLLExports int WINAPI Help( LPTSTR inString )
{
	if ( !_wcsicmp( inString, L"popupfix" ) ) {
		MessageBox( GetForegroundWindow(), PopupFixHelpText, HelpWindowTitle, MB_OK );
		return 1;
	}
	else if ( !_wcsicmp( inString, L"_conemu" ) ) {
		MessageBox( GetForegroundWindow(), _ConEmuHelpText, HelpWindowTitle, MB_OK );
		return 1;
	}
	else if ( !_wcsicmp( inString, L"_console2" ) ) {
		MessageBox( GetForegroundWindow(), _Console2HelpText, HelpWindowTitle, MB_OK );
		return 1;
	}
	else if ( !_wcsicmp( inString, L"_tcmdtab" ) ) {
		MessageBox( GetForegroundWindow(), _TCmdTabHelpText, HelpWindowTitle, MB_OK );
		return 1;
	}
	else if ( !_wcsicmp( inString, L"_winterm" ) ) {
		MessageBox( GetForegroundWindow(), _WinTermHelpText, HelpWindowTitle, MB_OK );
		return 1;
	}
	else
		return 0;
}



DLLExports int WINAPI Usage( LPTSTR inString, LPTSTR outString )
{
	if ( !_wcsicmp( inString, L"popupfix" ) ) {
		wcscpy_s( outString, SHORT_BUF_LEN, UsageText( PopupFixHelpText ) );
		return 1;
	}
	else
		return 0;
}



/*

	1.3.0.1		2023-07-27

	1.4.0.0		2023-11-28
				Added support for Windows Terminal.  Use ShowCmdHelp() to
				display help text.  CenterPopups defaults to CENTER_ALWAYS.
				Moved the arg parser out to ParseArgs.cpp.  Updated the
				plugin's web site.  Other minor code tweaks.

	1.4.0.1		2024-03-13
				Minor tweaks to ForceForegroundWindow().

	1.4.1.0		2024-04-24
				RESTORE_FOCUS_CONSOLE2 also refocuses in Windows Terminal.
				Don't keep rechecking Windows Terminal status unless we
				originally started in Windows Terminal.  Once in Windows
				Terminal, always in Windows Terminal -- although the window's
				handle might change.  Windows Terminal detection no longer
				relies on an environment variable.
				If running in Windows Terminal, don't bother trying to detect
				Take Command.
				In both Console2 and Window Terminal, periodically move the
				hidden (pseudo-) console window to coincide with the visible
				window (improves TCC's original positioning of popup windows.)
				The Console2Only registry value / environment variable now
				control plugin loading under both Console2 AND Windows Terminal.

	1.4.2.0		2024-04-30
				Creates a list of ancestral PIDs at startup, and uses this list
				when checking for Windows Terminal instead of calling
				GetParentPID() repeatedly.  (Not for Console2 at this time.)
				Windows Terminal checking is faster, and is done more
				frequently.

*/
